"""djLint function to call cssbeautifier."""

from __future__ import annotations

from functools import partial
from io import StringIO
from typing import TYPE_CHECKING

import cssbeautifier
import regex as re
from cssbeautifier.css.options import BeautifierOptions

from djlint.helpers import RE_FLAGS_IS, child_of_unformatted_block

if TYPE_CHECKING:
    from djlint.settings import Config


def format_css(html: str, config: Config) -> str:
    """Format css inside <style> tags."""

    def launch_formatter(
        config: Config, html: str, match: re.Match[str]
    ) -> str:
        """Add break after if not in ignored block."""
        if child_of_unformatted_block(config, html, match):
            return match.group()

        if not match.group(3).strip():
            return match.group()

        indent = len(match.group(1)) * " "

        # because of the param options for css-beautifier we cannot pass
        # in a fixed space leading.
        # so, call formatter twice, once with a fake indent.
        # check which lines changed (these are the formattable lines)
        # and add the leading space to them.

        config.css_config["indent_level"] = 1
        opts = BeautifierOptions(config.css_config)
        beautified_lines = cssbeautifier.beautify(
            match.group(3), opts
        ).splitlines()

        config.css_config["indent_level"] = 2
        opts = BeautifierOptions(config.css_config)
        beautified_lines_test = cssbeautifier.beautify(
            match.group(3), opts
        ).splitlines()

        with StringIO() as buf:
            for line, test in zip(beautified_lines, beautified_lines_test):
                buf.write("\n")
                if line != test:
                    buf.write(indent)
                buf.write(line)
            beautified = buf.getvalue()

        return match.group(1) + match.group(2) + beautified + "\n" + indent

    func = partial(launch_formatter, config, html)

    return re.sub(
        r"([ ]*?)(<(?:style)\b(?:\"[^\"]*\"|'[^']*'|{[^}]*}|[^'\">{}])*>)(.*?)(?=</style>)",
        func,
        html,
        flags=RE_FLAGS_IS,
    )



"""Collection of shared djLint functions."""
# ruff: noqa: ERA001

from __future__ import annotations

import itertools
from functools import cache
from typing import TYPE_CHECKING

import regex as re

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Final

    from typing_extensions import TypeVar

    from djlint.settings import Config

    T = TypeVar("T")

RE_FLAGS_IS: Final = re.I | re.S
RE_FLAGS_IX: Final = re.I | re.X
RE_FLAGS_MS: Final = re.M | re.S
RE_FLAGS_MX: Final = re.M | re.X
RE_FLAGS_IMS: Final = re.I | re.M | re.S
RE_FLAGS_IMX: Final = re.I | re.M | re.X
RE_FLAGS_ISX: Final = re.I | re.S | re.X
RE_FLAGS_IMSX: Final = re.I | re.M | re.S | re.X


def _last_item(iterable: Iterable[T], /) -> T | None:
    last = None
    for item in iterable:
        last = item
    return last


def is_ignored_block_opening(config: Config, item: str) -> bool:
    """Find ignored group opening.

    A valid ignored group opening tag will not be part of a
    single line block.
    """
    inline = _last_item(
        re.finditer(config.ignored_blocks_inline, item, flags=RE_FLAGS_IMSX)
    )
    last_index = (
        inline.end()  # get the last index. The ignored opening should start after this.
        if inline
        else 0
    )
    return bool(
        re.search(
            config.ignored_block_opening, item[last_index:], flags=RE_FLAGS_IX
        )
    )


def is_script_style_block_opening(config: Config, item: str) -> bool:
    """Find ignored group opening.

    A valid ignored group opening tag will not be part of a
    single line block.
    """
    inline = _last_item(
        re.finditer(config.script_style_inline, item, flags=RE_FLAGS_IMSX)
    )
    last_index = (
        inline.end()  # get the last index. The ignored opening should start after this.
        if inline
        else 0
    )
    return bool(
        re.search(
            config.script_style_opening, item[last_index:], flags=RE_FLAGS_IX
        )
    )


def inside_protected_trans_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Find ignored group closing.

    A valid ignored group closing tag will not be part of a
    single line block.

    True = non indentable > inside ignored trans block
    False = indentable > either inside a trans trimmed block, or somewhere else, but not a trans non trimmed :)
    """
    close_block = re.search(
        config.ignored_trans_blocks_closing, match.group(), flags=RE_FLAGS_IX
    )

    if not close_block:
        return False

    non_trimmed = _last_item(
        re.finditer(config.ignored_trans_blocks, html, flags=RE_FLAGS_ISX)
    )

    trimmed = _last_item(
        re.finditer(config.trans_trimmed_blocks, html, flags=RE_FLAGS_ISX)
    )

    # who is max?
    if non_trimmed and (not trimmed or non_trimmed.end() > trimmed.end()):
        # non trimmed!
        # check that this is not an inline block.
        non_trimmed_inline = bool(
            re.search(
                config.ignored_trans_blocks, match.group(), flags=RE_FLAGS_ISX
            )
        )

        if non_trimmed_inline:
            last_index = non_trimmed.end()  # get the last index. The ignored opening should start after this.

            return bool(
                re.search(
                    config.ignored_trans_blocks_closing,
                    html[last_index:],
                    flags=RE_FLAGS_IX,
                )
            )

        return close_block.end() <= non_trimmed.end()

    if trimmed:
        # inside a trimmed block, we can return true to continue as if
        # this is a indentable block
        return close_block.end() > trimmed.end()
    return False

    # print(close_block)
    # if non_trimmed:
    #     last_index = (
    #         non_trimmed[-1].end()
    #     )  # get the last index. The ignored opening should start after this.

    # return re.search(
    #     config.ignored_trans_blocks_closing,
    #     html[last_index:],
    #     flags=RE_FLAGS_IX,
    # )


def is_ignored_block_closing(config: Config, item: str) -> bool:
    """Find ignored group closing.

    A valid ignored group closing tag will not be part of a
    single line block.
    """
    inline = _last_item(
        re.finditer(config.ignored_inline_blocks, item, flags=RE_FLAGS_IX)
    )
    last_index = (
        inline.end()  # get the last index. The ignored opening should start after this.
        if inline
        else 0
    )
    return bool(
        re.search(
            config.ignored_block_closing, item[last_index:], flags=RE_FLAGS_IX
        )
    )


def is_script_style_block_closing(config: Config, item: str) -> bool:
    """Find ignored group closing.

    A valid ignored group closing tag will not be part of a
    single line block.
    """
    inline = _last_item(
        re.finditer(config.script_style_inline, item, flags=RE_FLAGS_IX)
    )
    last_index = (
        inline.end()  # get the last index. The ignored opening should start after this.
        if inline
        else 0
    )
    return bool(
        re.search(
            config.script_style_closing, item[last_index:], flags=RE_FLAGS_IX
        )
    )


def is_safe_closing_tag(config: Config, item: str) -> bool:
    """Find ignored group opening.

    A valid ignored group opening tag will not be part of a
    single line block.
    """
    inline = _last_item(
        re.finditer(
            config.ignored_inline_blocks + r" | " + config.ignored_blocks,
            item,
            flags=RE_FLAGS_IMSX,
        )
    )
    last_index = (
        inline.end()  # get the last index. The ignored opening should start after this.
        if inline
        else 0
    )
    return bool(
        re.search(config.safe_closing_tag, item[last_index:], flags=RE_FLAGS_IX)
    )


def inside_template_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Check if a re.Match is inside of a template block."""
    match_start, match_end = match.span()
    for ignored_match in re.finditer(
        config.template_blocks, html, flags=RE_FLAGS_IMSX
    ):
        ignored_match_start, ignored_match_end = ignored_match.span()
        if (
            ignored_match_start <= match_start
            and match_end <= ignored_match_end
        ):
            return True
    return False


@cache
def _inside_html_attribute(
    html: str, /, *, html_tag_regex: str
) -> tuple[tuple[int, int], ...]:
    return tuple(
        # group 3 are the attributes
        x.span(3)
        for x in re.finditer(html_tag_regex, html, flags=RE_FLAGS_IMSX)
    )


def inside_html_attribute(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Check if a re.Match is inside of an html attribute."""
    match_start, match_end = match.span()
    for ignored_match_start, ignored_match_end in _inside_html_attribute(
        html, html_tag_regex=config.html_tag_regex
    ):
        # span = (-1, -1) if no attributes are present
        if (
            ignored_match_start <= match_start
            and match_end <= ignored_match_end
        ):
            return True
    return False


def inside_ignored_linter_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Check if a re.Match is inside of a ignored linter block."""
    match_start, match_end = match.span()
    for ignored_match in re.finditer(
        config.ignored_linter_blocks, html, flags=RE_FLAGS_IMSX
    ):
        ignored_match_start, ignored_match_end = ignored_match.span()
        if (
            ignored_match_start <= match_start
            and match_end <= ignored_match_end
        ):
            return True
    return False


@cache
def _inside_ignored_block(
    html: str, /, *, ignored_blocks: str, ignored_inline_blocks: str
) -> tuple[tuple[int, int], ...]:
    return tuple(
        x.span()
        for x in itertools.chain(
            re.finditer(ignored_blocks, html, flags=RE_FLAGS_IMSX),
            re.finditer(ignored_inline_blocks, html, flags=RE_FLAGS_IX),
        )
    )


def inside_ignored_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    match_start, match_end = match.span()
    for ignored_match_start, ignored_match_end in _inside_ignored_block(
        html,
        ignored_blocks=config.ignored_blocks,
        ignored_inline_blocks=config.ignored_inline_blocks,
    ):
        if (
            ignored_match_start <= match_start
            and match_end <= ignored_match_end
        ):
            return True
    return False


@cache
def _child_of_unformatted_block(
    html: str, /, *, unformatted_blocks: str, unformatted_blocks_coarse: str
) -> tuple[tuple[int, int], ...]:
    if not re.search(unformatted_blocks_coarse, html, flags=RE_FLAGS_IMSX):
        return ()
    return tuple(
        x.span()
        for x in re.finditer(unformatted_blocks, html, flags=RE_FLAGS_IMSX)
    )


def child_of_unformatted_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    match_start, match_end = match.span()
    for ignored_match_start, ignored_match_end in _child_of_unformatted_block(
        html,
        unformatted_blocks=config.unformatted_blocks,
        unformatted_blocks_coarse=config.unformatted_blocks_coarse,
    ):
        if ignored_match_start < match_start and match_end <= ignored_match_end:
            return True
    return False


def child_of_ignored_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    match_start, match_end = match.span()
    for ignored_match in itertools.chain(
        re.finditer(config.ignored_blocks, html, flags=RE_FLAGS_IMSX),
        re.finditer(config.ignored_inline_blocks, html, flags=RE_FLAGS_IX),
    ):
        ignored_match_start, ignored_match_end = ignored_match.span()
        if ignored_match_start < match_start and match_end <= ignored_match_end:
            return True
    return False


@cache
def _overlaps_ignored_block(
    html: str, /, *, ignored_blocks: str, ignored_inline_blocks: str
) -> tuple[tuple[int, int], ...]:
    return tuple(
        x.span()
        for x in itertools.chain(
            re.finditer(ignored_blocks, html, flags=RE_FLAGS_IMSX),
            re.finditer(ignored_inline_blocks, html, flags=RE_FLAGS_IX),
        )
    )


def overlaps_ignored_block(
    config: Config, html: str, match: re.Match[str]
) -> bool:
    """Do not add whitespace if the tag is in a non indent block."""
    match_start, match_end = match.span()
    for ignored_match_start, ignored_match_end in _overlaps_ignored_block(
        html,
        ignored_blocks=config.ignored_blocks,
        ignored_inline_blocks=config.ignored_inline_blocks,
    ):
        # don't require the match to be fully inside the ignored block.
        # poorly build html will probably span ignored blocks and should be ignored.
        if (ignored_match_start <= match_start <= ignored_match_end) or (
            ignored_match_start <= match_end <= ignored_match_end
        ):
            return True
    return False


def inside_ignored_rule(
    config: Config, html: str, match: re.Match[str], rule: str
) -> bool:
    """Check if match is inside an ignored pattern."""
    match_start, match_end = match.span()
    for rule_regex in config.ignored_rules:
        for ignored_match in re.finditer(rule_regex, html, flags=RE_FLAGS_ISX):
            ignored_match_start, ignored_match_end = ignored_match.span()
            if (
                (ignored_match_start <= match_start <= ignored_match_end)
                and rule in re.split(r"\s|,", ignored_match.group(1).strip())
            ) or (
                (ignored_match_start <= match_end <= ignored_match_end)
                and not ignored_match.group(1).strip()
            ):
                return True
    return False



"""Build src file list."""

from __future__ import annotations

from typing import TYPE_CHECKING

import regex as re
from click import echo
from colorama import Fore

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from djlint.settings import Config


def get_src(src: Iterable[Path], config: Config) -> list[Path]:
    """Get source files."""
    paths = []
    for item in src:
        # normalize path

        normalized_item = item.resolve()

        if normalized_item.is_file():
            if no_pragma(config, normalized_item) and (
                not config.use_gitignore
                or not config.gitignore.match_file(normalized_item)
            ):
                paths.append(normalized_item)
            continue

        # remove leading . from extension
        extension = config.extension.removeprefix(".")

        paths.extend(
            x
            for x in normalized_item.glob(f"**/*.{extension}")
            if (
                not re.search(config.exclude, x.as_posix(), flags=re.X)
                and no_pragma(config, x)
                and (
                    not config.use_gitignore
                    or not config.gitignore.match_file(x)
                )
            )
        )

    if not paths:
        echo(Fore.BLUE + "No files to check! 😢")

    return paths


html_patterns = (r"<!--\s*djlint\:on\s*-->",)
django_jinja_patterns = (
    r"\{#\s*djlint\:on\s*#\}",
    r"\{%\s*comment\s*%\}\s*djlint\:on\s*\{%\s*endcomment\s*%\}",
)
nunjucks_patterns = (r"\{#\s*djlint\:on\s*#\}",)
handlebars_patterns = (r"\{\{!--\s*djlint\:on\s*--\}\}",)
golang_patterns = (r"\{\{-?\s*/\*\s*djlint\:on\s*\*/\s*-?\}\}",)


def no_pragma(config: Config, this_file: Path) -> bool:
    """Verify there is no pragma present."""
    if not config.require_pragma:
        return True

    with this_file.open(encoding="utf-8") as open_file:
        first_line = open_file.readline()

    pragma_patterns = {
        "html": html_patterns,
        "django": django_jinja_patterns + html_patterns,
        "jinja": django_jinja_patterns + html_patterns,
        "nunjucks": nunjucks_patterns + html_patterns,
        "handlebars": handlebars_patterns + html_patterns,
        "golang": golang_patterns + html_patterns,
        "angular": html_patterns,
        "all": django_jinja_patterns
        + nunjucks_patterns
        + handlebars_patterns
        + golang_patterns
        + html_patterns,
    }

    for pattern in pragma_patterns[config.profile]:
        if re.match(pattern, first_line):
            return True
    return False



from typing_extensions import TypedDict

class LintError(TypedDict):
    code: str
    line: str
    match: str
    message: str

class ProcessResult(TypedDict, total=False):
    format_message: dict[str, tuple[str, ...]]
    lint_message: dict[str, list[LintError]]
